import path from "node:path";
import { Glob } from "bun";
import YAML from "yaml";
import z from "zod";

const BindingFileSchema = z.object({
  $binding: z.literal("playbill"),
  id: z.string(),
  name: z.string().default("Unnamed playbill"),
  author: z.string().optional(),
  description: z.string().optional(),
  version: z.string(),
  files: z.array(z.string()).default(["**/*.yaml", "**/*.md"]),
});

type BindingFileContent = z.infer<typeof BindingFileSchema>;

export interface PlaybillBinding {
  _raw: BindingFileContent;
  bindingFilePath: string;
  includedFiles: string[];

  id: string;
  name: string;
  author: string;
  version: string;
  description: string;
}

export function parseBindingFile(
  text: string,
  filePath: string,
): PlaybillBinding {
  const parsed = YAML.parse(text);
  const binding = BindingFileSchema.parse(parsed);
  const fileGlobs = binding.files;
  const bindingFileDirname = filePath.split("/").slice(0, -1).join("/");

  const allFilePaths: string[] = [];

  for (const thisGlob of fileGlobs) {
    const glob = new Glob(thisGlob);

    const results = glob.scanSync({
      cwd: bindingFileDirname,
      absolute: true,
      followSymlinks: true,
      onlyFiles: true,
    });

    allFilePaths.push(...results);
  }

  const bindingFileAbsolutePath = path.resolve(filePath);
  const filteredFilePaths = allFilePaths.filter((filePath) => {
    return filePath !== bindingFileAbsolutePath;
  });

  return {
    _raw: binding,
    bindingFilePath: filePath,
    includedFiles: filteredFilePaths,

    id: binding.id,
    name: binding.name,
    author: binding.author ?? "Anonymous",
    description: binding.description ?? "",
    version: binding.version,
  };
}
