import { parseArgs } from "node:util";
import {
  type AnyResource,
  ValidatedPlaybillSchema,
  checkDirectives,
  getEmptyPlaybill,
} from "@proscenium/playbill";
import chalk from "chalk";
import { parseBindingFile } from "./binding";
import { loadResourceFile } from "./resource";
import { usage } from "./usage";

enum ExitCode {
  Success = 0,
  Error = 1,
}

class CLIError extends Error { }

async function loadFileOrFail(filePath: string): Promise<string> {
  const fileToLoad = Bun.file(filePath);
  const fileExists = await fileToLoad.exists();

  if (!fileExists) {
    throw new Error(`File not found: ${filePath}`);
  }

  return await fileToLoad.text();
}

async function main(): Promise<number> {
  // -- ARGUMENT PARSING -- //
  const { values: options, positionals: args } = parseArgs({
    args: Bun.argv.slice(2),
    options: {
      write: {
        short: "w",
        type: "boolean",
        default: false,
      },
      check: {
        short: "c",
        type: "boolean",
        default: false,
      },
      outfile: {
        short: "o",
        type: "string",
        default: "playbill.json",
      },
      help: {
        short: "h",
        default: false,
        type: "boolean",
      },
      verbose: {
        short: "v",
        default: false,
        type: "boolean",
      },
      minify: {
        short: "m",
        default: false,
        type: "boolean",
      },
    },
    strict: true,
    allowPositionals: true,
  });

  // -- HELP TEXT -- //
  if (options.help) {
    console.log(usage);
    return ExitCode.Success;
  }

  // -- ARG VALIDATION -- //
  if (options.check && options.write) {
    throw new CLIError("Cannot use --check and --write together");
  }

  const VERBOSE = options.verbose;

  /**
   * Log a message if the vervose flag has been set
   * @param msg - The message to log
   */
  function verboseLog(msg: string) {
    if (VERBOSE) {
      console.log(chalk.dim(msg));
    }
  }

  // -- BINDING FILE -- //
  const bindingPath = args[0] ?? "./binding.yaml";

  verboseLog(`Building Playbill with binding: ${bindingPath}`);

  const bindingFileContent = await loadFileOrFail(bindingPath);
  const binding = parseBindingFile(bindingFileContent, bindingPath);

  verboseLog(
    `↳ Binding loaded with ${binding.includedFiles.length} associated resources`,
  );

  // -- RESOURCE FILES -- //
  verboseLog("Loading resources");
  const loadedResources: AnyResource[] = [];
  for (const filepath of binding.includedFiles) {
    verboseLog(`↳ Loading ${filepath}`);
    const loaded = await loadResourceFile(filepath);
    verboseLog(`  ↳ Loaded ${loaded.length} resources`);
    loadedResources.push(...loaded);
  }

  verboseLog("Constructing playbill");

  // -- COMPILE RESOURCES --//
  const playbill = getEmptyPlaybill();

  playbill.id = binding.id;
  playbill.name = binding.name;
  playbill.author = binding.author;
  playbill.version = binding.version;
  playbill.description = binding.description;

  for (const resource of loadedResources) {
    switch (resource.$define) {
      case "ability":
        playbill.abilities.push(resource);
        break;
      case "species":
        playbill.species.push(resource);
        break;
      case "entity":
        playbill.entities.push(resource);
        break;
      case "item":
        playbill.items.push(resource);
        break;
      case "tag":
        playbill.tags.push(resource);
        break;
      case "method":
        playbill.methods.push(resource);
        break;
      case "lore":
        playbill.lore.push(resource);
        break;
      case "rule":
        playbill.rules.push(resource);
        break;
      case "playbill":
        throw new Error("Cannot load playbills rn dawg");
    }
  }

  // -- VALDATE PLAYBILL -- //
  verboseLog("Validating playbill");
  const validatedPlaybill = ValidatedPlaybillSchema.safeParse(playbill);

  if (!validatedPlaybill.success) {
    console.error("Error validating playbill");
    console.error(validatedPlaybill.error.errors);
    return ExitCode.Error;
  }

  const directiveValidationResult = checkDirectives(validatedPlaybill.data);

  if (!directiveValidationResult) {
    console.error("Error validating playbill directives");
    return ExitCode.Error;
  }

  verboseLog("Playbill validated");

  // -- EXIT EARLY IF JUST CHECKING VALIDATION --//
  if (options.check) {
    console.log(chalk.green("Playbill validated successfully"));
    return ExitCode.Error;
  }

  // -- SERIALIZE TO JSON --//
  const finalizedPlaybill = JSON.stringify(
    validatedPlaybill.data,
    null,
    options.minify ? undefined : 2,
  );

  // -- WRITE TO DISK OR STDOUT --//
  if (options.write) {
    await Bun.write(options.outfile, finalizedPlaybill);
    return ExitCode.Success;
  }

  console.log(finalizedPlaybill);
  return ExitCode.Success;
}

try {
  const exitCode = await main();
  process.exit(exitCode);
} catch (error) {
  if (error instanceof Error) {
    console.error("Uncaught error:");
    console.error(chalk.red(error.message));
    console.error(error.stack);
  } else {
    console.error("An unknown error occurred");
    console.error(error);
  }

  process.exit(1);
}
